#version 130

uniform sampler2D ImageSampler2D_u;
uniform sampler2D ImageSampler2D_v;

uniform sampler2D ImageSampler2D_left;
uniform sampler2D ImageSampler2D_right;

uniform float gamma;
uniform float scale;

uniform int shader_mode;




// from: https://love2d.org/wiki/HSV_color
vec3 hsv2rgb(float h, float s, float v)
{
  if (s <= 0 ) { return vec3 ( v ); }
  h = h / 60.0;
  float c = v * s;
  float x = ( 1 - abs( ( mod( h, 2 ) - 1 ) ) ) * c;
  float m = v - c;
  float r = 0.0;
  float g = 0.0;
  float b = 0.0;

  if (h < 1) { r = c; g = x;b = 0.0; }
  else if (h < 2) { r = x; g = c; b = 0.0; }
  else if (h < 3) { r = 0.0; g = c; b = x; }
  else if (h < 4) { r = 0.0; g = x; b = c; }
  else if (h < 5) { r = x; g = 0.0; b = c; }
  else  { r = c; g = 0.0; b = x; }

  return vec3( r + m, g + m, b + m );
}




void main() {

  float pi = 3.14159265358;

  // only flip y
  ivec2 texture_size= textureSize(ImageSampler2D_left, 0 );
  ivec2 texCoord = ivec2((ivec2(gl_FragCoord.x, texture_size.y) - ivec2(0,1) - ivec2(0,gl_FragCoord.y)));

	vec4 offx = texelFetch( ImageSampler2D_u, texCoord, 0);
	vec4 offy = texelFetch( ImageSampler2D_v, texCoord, 0);

	vec2 size = textureSize(ImageSampler2D_left, 0 );

        //vec4 ImgVal_warped = texelFetch( ImageSampler2D_right, texCoord + vec2( offx.x, offy.x ), 0);
	// need texture2D here for interpolated image lookup
        vec4 ImgVal_warped = texture2D ( ImageSampler2D_right, (texCoord + 2 * vec2( offx.x, offy.x ) + vec2(0.5)) / size , 0);
	vec4 ImgVal_left   = texelFetch( ImageSampler2D_left , texCoord, 0);
        vec4 ImgVal_right  = texelFetch( ImageSampler2D_right, texCoord, 0);


	float val;
	float u = offx.x;
	float v = offy.x;

	// how to do this better ?
	if ( shader_mode == 1 )
	  val = u;	  

	if ( shader_mode == 2 )
	  val = v;	  

	if ( shader_mode == 3 )
	  {
	    // Middlebury approximation
	    
	    float alpha = mod( atan( -v, -u ) / pi * 180 - 20, 360 ); 
	    float length = sqrt( u * u + v * v ) * scale;
	    float value;

	    if ( length > 1 )
	      value = 0.5;
	    else
	      value = 1.0;

	    gl_FragColor.rgb = hsv2rgb( alpha, length, value );
	  }
	if ( shader_mode == 4 )
	  {
	    /* IPOL approximation */
	   
	    float alpha = mod( atan( -v, -u ) / pi * 180, 360 ); 
	    float length = sqrt( u * u + v * v ) * scale;
	    float value;

	    if ( length > 1 )
	      value = 1.0;
	    else
	      value = length;

	    gl_FragColor.rgb = hsv2rgb( alpha, length, value );
	    

	  }


	if ( shader_mode == 5 )
	  {
	    gl_FragColor.r = pow( ImgVal_left.r, 1 / gamma );	  
	    gl_FragColor.g = pow( ImgVal_left.g, 1 / gamma );	  
	    gl_FragColor.b = pow( ImgVal_left.b, 1 / gamma );	  
	  }
	
	if ( shader_mode == 6 )
	  {
	    gl_FragColor.r = pow( ImgVal_right.r, 1 / gamma );	  
	    gl_FragColor.g = pow( ImgVal_right.g, 1 / gamma );	  
	    gl_FragColor.b = pow( ImgVal_right.b, 1 / gamma );	  
	  }

	if ( shader_mode == 7 )
	  {
	    gl_FragColor.r = pow( ImgVal_warped.r, 1 / gamma );	  
	    gl_FragColor.g = pow( ImgVal_warped.g, 1 / gamma );	  
	    gl_FragColor.b = pow( ImgVal_warped.b, 1 / gamma );	  
	  }



	if ( shader_mode == 1 ||
	     shader_mode == 2 )
	  {	     
	    if ( val < 0 )
	      gl_FragColor.r = pow( -val * scale, 1 / gamma );
	    if ( val >= 0 )		 
	      gl_FragColor.g = pow( val * scale, 1 / gamma );	
	    
	    if( isnan( val ) )
	      gl_FragColor.b = 1.0;

	    if( isinf( val ) )
	      gl_FragColor.b = 0.5;

	    
	  }
	gl_FragColor.a = 1.0;
}
