/*                                                                                                                                
   minexample.cu
               
   - assigns some numbers to an image downloads the result
     and saves it
                                                                                                                                  
   (c) 2012 Ivo Ihrke                                                                                                        
                                                                                                                                  
   Universitaet des Saarlandes / MPI Informatik                                                                                   
                                                                                                                                  
   ihrke@mmci.uni-saarland.de                                                                                                     
   
   
   thanks to an initial version 

   Copyright (C) 2010 Bastian Goldluecke,
                                                                                                                               
*/                                                                                                                                

#define MAIN_PROG

#include <iostream>
#include <stdio.h>
#include <assert.h>

using namespace std;


// Our image class
#include "../../Image/pvc_image.h"



#define cuda_float float

// constants for the width and height of our image
// #define W 256
// #define H 256

int W;
int H;

// CUDA Block size
#define BLOCK_SIZE_X 16
#define BLOCK_SIZE_Y 16

// Macro to handle CUDA error return
#define CUDA_SAFE_CALL(c) if ( c != cudaSuccess ) { cout << "CUDA error: " << cudaGetErrorString( cudaGetLastError() )  << ", line " << __LINE__ << endl; cout.flush(); assert( false ); }




// example kernel that reads input and modifies it
__global__ void multiply_input( cuda_float *px, int W ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  px [ off + 0 ] = px [ off + 0 ] * (cuda_float) 0.8;
  px [ off + 1 ] = px [ off + 1 ] * (cuda_float) 0.5;
  px [ off + 2 ] = px [ off + 2 ] * (cuda_float) 0.3;
  px [ off + 3 ] = px [ off + 3 ] * (cuda_float) 1.0;

     
  __syncthreads();
       

}



// example kernel that reads input
__global__ void gauss_analytic( cuda_float *px, cuda_float *out, int W, int halfsize, cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  


  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int i = -halfsize; i <= halfsize; i++ )
	for ( int j = -halfsize; j <= halfsize; j++ )
	  { 
	    int img_off = ( i + j * W ) * 4;

	    cuda_float img_val = px [ off + img_off + c ];

	    cuda_float square = i * i + j * j;
	    
	    cuda_float kval = exp( - square / ( 2.0 * SIGMA * SIGMA ) );

	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }
  
  __syncthreads();
       

}


// example kernel that reads input
__global__ void gauss_sep_hor( cuda_float *px, cuda_float *out, const int W, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int i = -halfsize; i <= halfsize; i++ )
	  { 
	    int img_off = i * 4;
	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = i * i;
	    
	    cuda_float kval = exp( - square * SIGMA_square );

	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }
  
  //  __syncthreads();
       

}


// example kernel that reads input
__global__ void gauss_sep_ver( cuda_float *px, cuda_float *out, const int W, const int halfsize, const cuda_float SIGMA ) 
  {
  // Global thread location
  int ox = BLOCK_SIZE_X * blockIdx.x + threadIdx.x;
  int oy = BLOCK_SIZE_Y * blockIdx.y + threadIdx.y;

  int off = ( ox + oy * W ) * 4;
  
  cuda_float SIGMA_square = 1.0f / ( SIGMA * SIGMA * 2.0f );

  for ( int c = 0; c < 4; c++ )
    {

      cuda_float sum    = (cuda_float) 0.0;
      cuda_float weight = (cuda_float) 0.0;

      for ( int j = -halfsize; j <= halfsize; j++ )
	  { 
	    int img_off = j * W * 4;
	    
	    cuda_float img_val = px [ off + img_off + c ];
	    
	    cuda_float square = j * j;
	    
	    //	    cuda_float kval = exp( - square / ( 2.0 * SIGMA * SIGMA ) );
	    cuda_float kval = exp( - square * SIGMA_square );


	    weight = weight + kval;

	    sum = sum + img_val * kval;
	  }
      
      out [ off + c ] = sum / weight;
    }
  
  //__syncthreads();
       

}



// from CUDA SDK
// Beginning of GPU Architecture definitions
inline int _ConvertSMVer2Cores(int major, int minor)
{
  // Defines for GPU Architecture types (using the SM version to determine the # of cores per SM
  typedef struct {
    int SM; // 0xMm (hexidecimal notation), M = SM Major version, and m = SM minor version
    int Cores;
  } sSMtoCores;

  sSMtoCores nGpuArchCoresPerSM[] = 
    { { 0x10,  8 },
      { 0x11,  8 },
      { 0x12,  8 },
      { 0x13,  8 },
      { 0x20, 32 },
      { 0x21, 48 },
      {   -1, -1 } 
    };

  int index = 0;
  while (nGpuArchCoresPerSM[index].SM != -1) {
    if (nGpuArchCoresPerSM[index].SM == ((major << 4) + minor) ) {
      return nGpuArchCoresPerSM[index].Cores;
    }
    index++;
  }
  printf("MapSMtoCores undefined SMversion %d.%d!\n", major, minor);
  return -1;
}
// end of GPU Architecture definitions




// this one checks if your card is suitable and drivers are working
int init_CUDA()
{
  // Check available devices
  int error = 0;
  int deviceCount;
  int gpuDeviceCount = 0;
  struct cudaDeviceProp deviceProp;
  
  cudaError_t result = cudaGetDeviceCount( &deviceCount );

  if ( result != cudaSuccess )
    {
      cout << "No CUDA devices found - Error string: " << cudaGetErrorString( result ) << endl;
      
      // machines without GPU should still support emulation
      deviceCount = 0; 
      
      error = -1;
    }
  
  for ( int device = 0; device < deviceCount; ++device ) 
    {
      CUDA_SAFE_CALL ( cudaGetDeviceProperties(&deviceProp, device) );
      
      if ( deviceProp.major != 9999) /* 9999 means emulation only */
	++gpuDeviceCount;
    }
  
  cout <<  gpuDeviceCount << " GPU CUDA device(s) found" << endl;

  // Use first device found
  CUDA_SAFE_CALL ( cudaSetDevice( 0 ) );


  cudaGetDeviceProperties( &deviceProp, 0 );
  cout << "Using CUDA device " << deviceProp.name << endl;

  printf("[%s] has %d MP(s) x %d (Cores/MP) = %d (Cores)\n",
	 deviceProp.name, deviceProp.multiProcessorCount,
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor),
	 _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor) * deviceProp.multiProcessorCount);

  cout << "ClockSpeed: " << deviceProp.clockRate << " Hz (?)." << endl;

  cout.flush();

  return error;
}








  // --------------------------------------------------------------------------------
  // ------------------------------------- main function ----------------------------
  // --------------------------------------------------------------------------------

// Main function for example
int main( int argc, char *argv [] )
{
  
  
  // this is where our result will be stored
  PVC_Image  *result = NULL;
  PVC_Image  *input  = NULL;

  // a CUDA GPU-based array 
  cuda_float *gpu_px  = NULL;
  cuda_float *gpu_dum = NULL;
  cuda_float *gpu_out = NULL;

  // report errors in this variable
  cudaError_t err;

  
  // ------------ *NEW* begin --------------
  input = new PVC_Image;
  //input -> load("../../Data/earth_small.ppm");
  //  input -> load("../../Data/depth highres.ppm");
  input -> load("../../Data/earth.ppm");

  // number of pixels
  W = input -> w();
  H = input -> h();
  int N = W * H;

  // ------------ *NEW* end ----------------



  if ( init_CUDA() < 0 )
    {
      fprintf( stderr,"Could not initialize CUDA, exiting ...\n" );
      exit ( -1 );
    }


  // allocate N * 4 floats on the GPU (4 color channels) - will be the input now
  CUDA_SAFE_CALL( cudaMalloc( &gpu_px, N * sizeof( cuda_float ) * 4 ) );

  // ------------ *NEW* begin --------------x
  // allocate output 
  CUDA_SAFE_CALL( cudaMalloc( &gpu_dum, N * sizeof( cuda_float ) * 4 ) );
  CUDA_SAFE_CALL( cudaMalloc( &gpu_out, N * sizeof( cuda_float ) * 4 ) );
  // ------------ *NEW* end   --------------

  
  // initialize our result image
  result = new PVC_Image;
  result -> resize( W, H );
  
  // ------------ *NEW* begin --------------

  // Copy matrices to GPU (host to device)
  CUDA_SAFE_CALL( cudaMemcpy( gpu_px,     input -> get_data(), N * sizeof(cuda_float) * 4, cudaMemcpyHostToDevice ));
  // ------------ *NEW* end ----------------

  

  // Set block size (in number of threads)
  dim3 dimBlock( BLOCK_SIZE_X, BLOCK_SIZE_Y );

  // Set grid size (in number of blocks)
  dim3 dimGrid( W / dimBlock.x, H / dimBlock.y ); 

  // Perform computation
  cout << "GPU computation: ["; cout.flush();
  clock_t gpu_start = clock();


  
  // issue the computation
  //  multiply_input <<< dimGrid, dimBlock >>> ( gpu_px, W );
  //  gauss_analytic <<< dimGrid, dimBlock >>> ( gpu_px, gpu_out, W, 30, 15.0 );
  gauss_sep_hor <<< dimGrid, dimBlock >>> ( gpu_px, gpu_dum, W, 100, 50.0 );

  // make sure all threads are done
  cudaThreadSynchronize();

  gauss_sep_ver <<< dimGrid, dimBlock >>> ( gpu_dum, gpu_out, W, 100, 50.0 );

  // make sure all threads are done
  //cudaThreadSynchronize();

  // did it work ?
  err = cudaGetLastError();
  if (err != cudaSuccess) 
    printf("Error: %s\n", cudaGetErrorString(err));  

  // timing
  clock_t gpu_end = clock();

  cout << "] done." << endl;
  cout << "GPU time: " << (gpu_end - gpu_start) / double(CLOCKS_PER_SEC) << endl;  
  

  // Copy back result to CPU
  // we can read directly to our image class because 
  // we have non-padded memory (just W*H floats)

  CUDA_SAFE_CALL( cudaMemcpy( result -> get_data()  , gpu_out , N * sizeof(cuda_float) * 4, cudaMemcpyDeviceToHost ) );


  //for( int i = 0; i < N * 4; i++ )
  //  result -> get_data() [ i ] = input -> get_data() [ i ];

  // Save results

  result -> save( "Output/gauss.ppm" );

  // Cleanup
  // GPU
  CUDA_SAFE_CALL( cudaFree( gpu_px ));

  // ------------ *NEW* begin --------------
  CUDA_SAFE_CALL( cudaFree( gpu_dum ));
  CUDA_SAFE_CALL( cudaFree( gpu_out ));
  // ------------ *NEW* end ----------------

  // Host
  delete result;
  delete input;

  return 0;
}

